@@ -80,7 +80,7 @@ let max_command_length =
 let get_output ?(use_bash = false) ?(accept_error=false) cmd =
   let fn = Filename.temp_file "flexdll" "" in
   let cmd' = cmd ^ " > " ^ (Filename.quote fn) in
-    if String.length cmd' < max_command_length && not use_bash then
+    if String.length cmd' <= max_command_length && not use_bash then
       begin
         if (Sys.command cmd' <> 0) && not accept_error
         then failwith ("Cannot run " ^ cmd);
@@ -202,27 +202,35 @@ let new_cmdline () =
   }
 
 let run_command cmdline cmd =
-  let cmd_quiet =
-    match !toolchain with
-    | `MSVC | `MSVC64 when !verbose < 1 -> cmd ^ " >NUL"
-    | _ -> cmd
+  let pipe_to_null = (!toolchain = `MSVC || !toolchain = `MSVC64) in
+  let silencer = if pipe_to_null then " >NUL 2>NUL" else ""
   in
   (* note: for Cygwin, using bash allow to follow symlinks to find
      gcc... *)
   if !toolchain = `CYGWIN || !toolchain = `CYGWIN64 ||
-     String.length cmd_quiet >= max_command_length
+     String.length cmd + String.length silencer > max_command_length
   then begin
     (* Dump the command in a text file and apply bash to it. *)
     let (fn, oc) = open_temp_file "longcmd" "" in
     output_string oc cmd;
+    if pipe_to_null then
+      output_string oc " &>/dev/null";
     close_out oc;
 
-    debug 1 "(call with bash: %s)" fn;
-    if Sys.command (Printf.sprintf "bash %s" fn) <> 0 then
+    debug 1 "(call with bash: %s)\n%!" fn;
+    let invoke = Printf.sprintf "bash %s" fn in
+    if Sys.command invoke <> 0 then begin
+      if pipe_to_null then begin
+        let oc = open_out fn in
+        output_string oc cmd;
+        close_out oc;
+        ignore (Sys.command invoke)
+      end;
       failwith "Error during linking\n"
+    end
   end else
-    if Sys.command cmd_quiet <> 0 then begin
-      if cmd <> cmd_quiet then ignore (Sys.command cmd);
+    if Sys.command (cmd ^ silencer) <> 0 then begin
+      if pipe_to_null then ignore (Sys.command cmd);
       failwith "Error during linking\n"
     end
 
@@ -1041,6 +1049,15 @@ let build_dll link_exe output_file files exts extra_args =
         let extra_args =
           if !machine = `x64 then (Printf.sprintf "/base:%s " !base_addr) ^ extra_args else extra_args
         in
+
+        let extra_args =
+          (* FlexDLL doesn't process .voltbl sections correctly, so don't allow the linker
+             to process them. *)
+          if Sys.command "link | findstr EMITVOLATILEMETADATA > nul" = 0 then
+            "/EMITVOLATILEMETADATA:NO " ^ extra_args
+          else extra_args
+        in
+
         (* Flexdll requires that all images (main programs and all the DLLs) are
            not too far away. This is needed because of the 32-bit relative relocations
            (e.g. function calls). It seems that passing such a /base argument to link.exe
@@ -1075,9 +1092,6 @@ let build_dll link_exe output_file files exts extra_args =
             close_out oc;
             Filename.quote def_file
         in
-        let extra_args =
-          if !machine = `x64 then "-Xlinker --image-base -Xlinker 0x10000 " ^ extra_args else extra_args
-        in
         Printf.sprintf
           "%s %s%s -L. %s %s -o %s %s %s %s %s"
           !gcc
@@ -1203,7 +1217,11 @@ let normalize_path path =
   let rec loop acc path =
     match path with
     | "."  :: path -> loop acc path
-    | ".." :: path -> loop (List.tl acc) path
+    | ".." :: path -> begin
+        match acc with
+        | ".." :: _ | [] -> loop (".." :: acc) path
+        | _ :: acc -> loop acc path
+      end
     | elem :: path -> loop (elem :: acc) path
     | [] -> List.rev acc
   in
