@@ -130,6 +130,14 @@ let family =
   ) in
   fun () -> Lazy.force family
 
+let yum_cmd = lazy begin
+  if OpamSystem.resolve_command "yum" <> None then
+    "yum"
+  else if OpamSystem.resolve_command "dnf" <> None then
+    "dnf"
+  else
+    raise (OpamSystem.Command_not_found "yum or dnf")
+end
 
 let packages_status packages =
   let (+++) pkg set = OpamSysPkg.Set.add (OpamSysPkg.of_string pkg) set in
@@ -185,11 +193,17 @@ let packages_status packages =
         let short_name =
           match String.rindex pkg '/' with
           | exception Not_found -> pkg
-          | idx -> String.sub pkg idx (String.length pkg - idx)
+          | idx -> String.sub pkg (idx+1) (String.length pkg - idx - 1)
+        in
+        let no_flavor =
+          match String.index short_name ',' with
+          | exception Not_found -> short_name
+          | idx -> String.sub short_name 0 idx
         in
         set
         |> OpamSysPkg.Set.add (OpamSysPkg.of_string pkg)
         |> OpamSysPkg.Set.add (OpamSysPkg.of_string short_name)
+        |> OpamSysPkg.Set.add (OpamSysPkg.of_string no_flavor)
       ) OpamSysPkg.Set.empty l
   in
   let compute_sets_with_virtual get_avail_w_virtuals get_installed  =
@@ -635,11 +649,11 @@ let install_packages_commands_t sys_packages =
     let epel_release = "epel-release" in
     let install_epel rest =
       if List.mem epel_release packages then
-        ["yum", "install"::yes ["-y"] [epel_release]] @ rest
+        [Lazy.force yum_cmd, "install"::yes ["-y"] [epel_release]] @ rest
       else rest
     in
     install_epel
-      ["yum", "install"::yes ["-y"]
+      [Lazy.force yum_cmd, "install"::yes ["-y"]
                 (OpamStd.String.Set.of_list packages
                  |> OpamStd.String.Set.remove epel_release
                  |> OpamStd.String.Set.elements);
@@ -704,12 +718,12 @@ let update () =
     match family () with
     | Alpine -> Some ("apk", ["update"])
     | Arch -> Some ("pacman", ["-Sy"])
-    | Centos -> Some ("yum", ["makecache"])
+    | Centos -> Some (Lazy.force yum_cmd, ["makecache"])
     | Debian -> Some ("apt-get", ["update"])
     | Gentoo -> Some ("emerge", ["--sync"])
     | Homebrew -> Some ("brew", ["update"])
     | Macports -> Some ("port", ["sync"])
-    | Suse -> Some ("zypper", ["--non-interactive"; "update"])
+    | Suse -> Some ("zypper", ["--non-interactive"; "refresh"])
     | Freebsd | Netbsd | Openbsd ->
       None
   in
