@@ -1,6 +1,6 @@
 (* TEST
 
-flags = "-w A"
+flags = "-w +A-70"
 
 * setup-ocamlc.byte-build-env
 ** ocamlc.byte
@@ -23,5 +23,5 @@ let h R.{x} = R.{x}
 open N (* used open *)
 let i (A|B) = B
 
-open! M (* open! also deactivates unused open warning *)
+open! M (* useless open! *)
 open M (* useless open *)
